/*	$NetBSD$	*/

/*
 * Copyright (c) 2003 Shigeyuki Fukushima.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Christopher G. Demetriou
 *	for the NetBSD Project.
 * 4. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "locators.h"

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/device.h>

#include <machine/bus.h>

#include <powerpc/ibm4xx/ibm405gp.h>
#include <powerpc/ibm4xx/dev/opbvar.h>
#include <powerpc/ibm4xx/dev/iicreg.h>

/*
 * Software state per device.
 */
struct iic_softc {
	struct device		sc_dev;		/* generic device information */
	bus_space_tag_t		sc_st;		/* bus space tag */
	bus_space_handle_t	sc_sh;		/* bus space handle */
	bus_dma_tag_t		sc_dmat;	/* bus DMA tag */
};

#define IIC_READ(sc, reg) \
	bus_space_read_1(sc->sc_st, sc->sc_sh, (reg), (val))
#define IIC_WRITE(sc, reg, val) \
	bus_space_write_1(sc->sc_st, sc->sc_sh, (reg), (val))

static int	iic_match(struct device *, struct cfdata *, void *);
static void	iic_attach(struct device *, struct device *, void *);

static void	iic_reset(struct iic_softc *sc);

CFATTACH_DECL(iic, sizeof(struct device),
    iic_match, iic_attach, NULL, NULL);

/*
 * Probe for the IIC.
 */
static int
iic_match(struct device *parent, struct cfdata *cf, void *aux)
{
	struct opb_attach_args *oaa = aux;

	/* match only on-chip iic devices */
	if (strcmp(oaa->opb_name, cf->cf_name) == 0)
		return (1);

	return (0);
}

/*
 * Attach the IIC.
 */
static void
iic_attach(struct device *parent, struct device *self, void *aux)
{
	struct opb_attach_args *oaa = aux;
	struct iic_softc *sc = (struct iic_softc *)self;
	/* int i; */

	sc->sc_st = oaa->opb_bt;
	sc->sc_sh = oaa->opb_addr;
	sc->sc_dmat = oaa->opb_dmat;

	printf(": 405GP IIC\n");

	/*
	 * Reset the chip to a known state.
	 */
	iic_reset();

	return;
}

static void
iic_reset(struct iic_softc *sc)
{
	/* Clear Low/High Master Address */
	IIC_WRITE(sc, IIC_LMADR, 0x0);
	IIC_WRITE(sc, IIC_HMADR, 0x0);

	/* Clear Low/High Slave Address */
	IIC_WRITE(sc, IIC_LSADR, 0x0);
	IIC_WRITE(sc, IIC_HSADR, 0x0);

	/* Clear any active interrupts */
	IIC_WRITE(sc, IIC_STS, IIC_STS_SCMP | IIC_STS_IRQA);

	/* Clear all writable status bits */
	IIC_WRITE(sc, IIC_EXTSTS, IIC_EXTSTS_IRQP
	 | IIC_EXTSTS_IRQD
	 | IIC_EXTSTS_LA
	 | IIC_EXTSTS_ICT
	 | IIC_EXTSTS_XFRA);

	/* Clear all writable status bits */
	IIC_WRITE(sc, IIC_CLKDIV, IIC_CLKDIV_30MHZ); /* XXX: OBS266 33MHz */

	/* Clear all interrupt mask bits */
	IIC_WRITE(sc, IIC_INTRMSK, 0x0);

	/* Clear Transfer Counter */
	IIC_WRITE(sc, IIC_XFRCNT, 0x0);

	/* Clear status set */
	IIC_WRITE(sc, IIC_XTCNTLSS, IIC_XTCNTLSS_SRC
	 | IIC_XTCNTLSS_SRS
	 | IIC_XTCNTLSS_SWC
	 | IIC_XTCNTLSS_SWS);

	/*
	 * Initialize Mode Control
	 *   Flushes master and slave data buffers
	 *   Enables IIC Standard mode (100Kb clock)
	 *   Disables Interrupts
	 *   Disables Slave Mode
	 *   Enables Exit Unknown State
	 */
	IIC_WRITE(sc, IIC_MDCNTL, IIC_MDCNTL_FSDB
	 | IIC_MDCNTL_FSMB
	 | IIC_MDCNTL_EUBS);

	IIC_WRITE(sc, IIC_CNTL, 0x0);

	return;
}
